/*	Copyright: 	 Copyright 2002 Apple Computer, Inc. All rights reserved.

	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
			("Apple") in consideration of your agreement to the following terms, and your
			use, installation, modification or redistribution of this Apple software
			constitutes acceptance of these terms.  If you do not agree with these terms,
			please do not use, install, modify or redistribute this Apple software.

			In consideration of your agreement to abide by the following terms, and subject
			to these terms, Apple grants you a personal, non-exclusive license, under Apples
			copyrights in this original Apple software (the "Apple Software"), to use,
			reproduce, modify and redistribute the Apple Software, with or without
			modifications, in source and/or binary forms; provided that if you redistribute
			the Apple Software in its entirety and without modifications, you must retain
			this notice and the following text and disclaimers in all such redistributions of
			the Apple Software.  Neither the name, trademarks, service marks or logos of
			Apple Computer, Inc. may be used to endorse or promote products derived from the
			Apple Software without specific prior written permission from Apple.  Except as
			expressly stated in this notice, no other rights or licenses, express or implied,
			are granted by Apple herein, including but not limited to any patent rights that
			may be infringed by your derivative works or by other works in which the Apple
			Software may be incorporated.

			The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
			WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
			WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
			PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
			COMBINATION WITH YOUR PRODUCTS.

			IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
			CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
			GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
			ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
			OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
			(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
			ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*=============================================================================
	AUScopeElement.cpp
	
=============================================================================*/

#include "AUScopeElement.h"
#include "AUBase.h"


//_____________________________________________________________________________
//
//	caller assumes that this is actually an immediate parameter
//
Float32			AUElement::GetParameter(AudioUnitParameterID paramID)
{
	ParameterMap::iterator i = mParameters.find(paramID);
	if (i == mParameters.end())
		COMPONENT_THROW(kAudioUnitErr_InvalidParameter);
		
	return (*i).second.GetValue();
}


//_____________________________________________________________________________
//
void			AUElement::GetRampSliceStartEnd(	AudioUnitParameterID paramID,
													Float32 &outStartValue,
													Float32 &outEndValue,
													Float32 &outValuePerFrameDelta )

{
	ParameterMap::iterator i = mParameters.find(paramID);
	if (i == mParameters.end())
		COMPONENT_THROW(kAudioUnitErr_InvalidParameter);

	ParameterMapEvent &event = (*i).second;
		
	// works even if the value is constant (immediate parameter value)
	event.GetRampSliceStartEnd(outStartValue, outEndValue, outValuePerFrameDelta );
}

//_____________________________________________________________________________
//
void			AUElement::SetParameter(AudioUnitParameterID paramID, Float32 inValue)
{
	ParameterMap::iterator i = mParameters.find(paramID);

	if (i == mParameters.end())
	{
		// create new entry in map for the paramID (only happens first time)
		ParameterMapEvent event(inValue);		
		mParameters[paramID] = event;
	}
	else
	{
		// paramID already exists in map so simply change its value
		ParameterMapEvent &event = (*i).second;
		event.SetValue(inValue);
	}
}

//_____________________________________________________________________________
//
void			AUElement::SetScheduledEvent(	AudioUnitParameterID 			paramID,
												const AudioUnitParameterEvent 	&inEvent,
												UInt32 							inSliceOffsetInBuffer,
												UInt32							inSliceDurationFrames )
{
	ParameterMap::iterator i = mParameters.find(paramID);

	if (i == mParameters.end())
	{
		// create new entry in map for the paramID (only happens first time)
		ParameterMapEvent event(inEvent, inSliceOffsetInBuffer, inSliceDurationFrames );
		mParameters[paramID] = event;
	}
	else
	{
		// paramID already exists in map so simply change its value
		ParameterMapEvent &event = (*i).second;
		
		event.SetScheduledEvent(inEvent, inSliceOffsetInBuffer, inSliceDurationFrames );
	}
}



//_____________________________________________________________________________
//
void			AUElement::GetParameterList(AudioUnitParameterID *outList)
{
	for (ParameterMap::iterator i = mParameters.begin(); i != mParameters.end(); ++i)
		*outList++ = (*i).first;
}

#if TARGET_API_MAC_OSX
#include <CoreFoundation/CFByteOrder.h>

//_____________________________________________________________________________
//
void			AUElement::SaveState(CFMutableDataRef data)
{
	UInt32 nparams = CFSwapInt32HostToBig(mParameters.size());
	CFDataAppendBytes(data, (UInt8 *)&nparams, sizeof(nparams));

	for (ParameterMap::iterator i = mParameters.begin(); i != mParameters.end(); ++i) {
		struct {
			UInt32				paramID;
			//CFSwappedFloat32	value; crashes gcc3 PFE
			UInt32				value;	// really a big-endian float
		} entry;
		
		entry.paramID = CFSwapInt32HostToBig((*i).first);

		Float32 v = (*i).second.GetValue();
		entry.value = CFSwapInt32HostToBig(*(UInt32 *)&v );

		CFDataAppendBytes(data, (UInt8 *)&entry, sizeof(entry));
	}
}

//_____________________________________________________________________________
//
const UInt8 *	AUElement::RestoreState(const UInt8 *state)
{
	const UInt8 *p = state;
	UInt32 nparams = CFSwapInt32BigToHost(*(UInt32 *)p);
	p += sizeof(UInt32);
	
	for (UInt32 i = 0; i < nparams; ++i) {
		struct {
			UInt32				paramID;
			Float32				value;
		} entry;
		
		entry.paramID = CFSwapInt32BigToHost(*(UInt32 *)p);
		p += sizeof(UInt32);
		//entry.value = CFConvertFloat32SwappedToHost(*(CFSwappedFloat32 *)p); 
		UInt32 temp = CFSwapInt32BigToHost(*(UInt32 *)p);
		entry.value = *(Float32 *)&temp;
		p += sizeof(CFSwappedFloat32);
		
		SetParameter(entry.paramID, entry.value);
	}
	return p;
}
#endif

//_____________________________________________________________________________
//
AUIOElement::AUIOElement(AUBase *audioUnit) :
	AUElement(audioUnit)
{
	mStreamFormat.SetCanonical(2,	// stereo
		audioUnit->AudioUnitAPIVersion() == 1);
		// interleaved if API version 1, deinterleaved if version 2
	mStreamFormat.mSampleRate = kAUDefaultSampleRate;
}

//_____________________________________________________________________________
//
OSStatus		AUIOElement::SetStreamFormat(const CAStreamBasicDescription &desc)
{
	mStreamFormat = desc;
	return noErr;
}

//_____________________________________________________________________________
//
void			AUIOElement::AllocateBuffer()
{
	if (GetAudioUnit()->IsInitialized())
		mIOBuffer.Allocate(mStreamFormat, 
			NeedsBufferSpace() ? GetAudioUnit()->GetMaxFramesPerSlice() : 0);
}

//_____________________________________________________________________________
//
void			AUIOElement::DeallocateBuffer()
{
	mIOBuffer.Deallocate();
}

//_____________________________________________________________________________
//
AUScope::~AUScope()
{
	for (std::vector<AUElement *>::iterator it = mElements.begin(); it != mElements.end(); ++it)
		delete *it;
}

//_____________________________________________________________________________
//
void	AUScope::SetNumberOfElements(UInt32 numElements)
{
	if (numElements > mElements.size()) {
		mElements.reserve(numElements);
		while (numElements > mElements.size()) {
			AUElement *elem = mCreator->CreateElement(mScope, mElements.size());
			mElements.push_back(elem);
		}
	} else
		while (numElements < mElements.size()) {
			AUElement *elem = mElements.back();
			mElements.pop_back();
			delete elem;
		}
}
