/*	Copyright: 	 Copyright 2002 Apple Computer, Inc. All rights reserved.

	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
			("Apple") in consideration of your agreement to the following terms, and your
			use, installation, modification or redistribution of this Apple software
			constitutes acceptance of these terms.  If you do not agree with these terms,
			please do not use, install, modify or redistribute this Apple software.

			In consideration of your agreement to abide by the following terms, and subject
			to these terms, Apple grants you a personal, non-exclusive license, under Apples
			copyrights in this original Apple software (the "Apple Software"), to use,
			reproduce, modify and redistribute the Apple Software, with or without
			modifications, in source and/or binary forms; provided that if you redistribute
			the Apple Software in its entirety and without modifications, you must retain
			this notice and the following text and disclaimers in all such redistributions of
			the Apple Software.  Neither the name, trademarks, service marks or logos of
			Apple Computer, Inc. may be used to endorse or promote products derived from the
			Apple Software without specific prior written permission from Apple.  Except as
			expressly stated in this notice, no other rights or licenses, express or implied,
			are granted by Apple herein, including but not limited to any patent rights that
			may be infringed by your derivative works or by other works in which the Apple
			Software may be incorporated.

			The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
			WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
			WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
			PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
			COMBINATION WITH YOUR PRODUCTS.

			IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
			CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
			GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
			ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
			OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
			(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
			ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*=============================================================================
	AUEffectBase.cpp
	
=============================================================================*/

#include "AUEffectBase.h"

AUEffectBase::AUEffectBase(	AudioUnit		audioUnit) :
	AUBase(audioUnit, 1, 1),		// 1 in bus, 1 out bus
	mBypassEffect(false),
	mParamSRDep (false)
{
}

AUEffectBase::~AUEffectBase()
{
	for (KernelList::iterator it = mKernelList.begin(); it != mKernelList.end(); ++it)
		delete *it;
}

ComponentResult		AUEffectBase::Initialize()
{
	MaintainKernels();
	return noErr;
}

ComponentResult		AUEffectBase::Reset(		AudioUnitScope 		inScope,
								 				AudioUnitElement 	inElement)
{
	for (KernelList::iterator it = mKernelList.begin(); it != mKernelList.end(); ++it) {
		AUKernelBase *kernel = *it;
		if (kernel != NULL)
			kernel->Reset();
	}
	
	return noErr;
}

ComponentResult		AUEffectBase::GetPropertyInfo (AudioUnitPropertyID	inID,
												AudioUnitScope					inScope,
												AudioUnitElement				inElement,
												UInt32 &						outDataSize,
												Boolean &						outWritable)
{
	if (inScope == kAudioUnitScope_Global) {
		switch (inID) {
			case kAudioUnitProperty_BypassEffect:
				outWritable = true;
				outDataSize = sizeof (UInt32);
				return noErr;
		}
	}
	return AUBase::GetPropertyInfo (inID, inScope, inElement, outDataSize, outWritable);
}


ComponentResult		AUEffectBase::GetProperty (AudioUnitPropertyID 		inID,
									  AudioUnitScope 					inScope,
									  AudioUnitElement			 		inElement,
									  void *							outData)
{
	if (inScope == kAudioUnitScope_Global) {
		switch (inID) {
			case kAudioUnitProperty_BypassEffect:
				*((UInt32*)outData) = (mBypassEffect ? 1 : 0);
				return noErr;
		}
	}
	return AUBase::GetProperty (inID, inScope, inElement, outData);
}


ComponentResult		AUEffectBase::SetProperty(		AudioUnitPropertyID inID,
									   AudioUnitScope 		inScope,
									   AudioUnitElement 	inElement,
									   const void *			inData,
									   UInt32 				inDataSize)
{
	if (inScope == kAudioUnitScope_Global) {
		switch (inID) {
			case kAudioUnitProperty_BypassEffect:
				bool tempNewSetting = *((UInt32*)inData) != 0;
				if (tempNewSetting && mBypassEffect == false) // turning bypass on
					Reset(0, 0);
				mBypassEffect = tempNewSetting;
				return noErr;
		}
	}
	return AUBase::SetProperty (inID, inScope, inElement, inData, inDataSize);
}
 

void	AUEffectBase::MaintainKernels()
{
	UInt32 nChannels = GetNumberOfChannels();
	
	if (mKernelList.size() < nChannels) {
		mKernelList.reserve(nChannels);
		for (UInt32 i = mKernelList.size(); i < nChannels; ++i)
			mKernelList.push_back(NewKernel());
	} else
		while (mKernelList.size() > nChannels) {
			AUKernelBase *kernel = mKernelList.back();
			delete kernel;
			mKernelList.pop_back();
		}

	for(unsigned int i = 0; i < nChannels; i++ )
	{
		mKernelList[i]->SetLastKernel(i == nChannels-1 );
	}
}

bool		AUEffectBase::StreamFormatWritable(	AudioUnitScope					scope,
												AudioUnitElement				element)
{
	return true;
}

ComponentResult		AUEffectBase::ChangeStreamFormat(	AudioUnitScope				inScope,
														AudioUnitElement			inElement,
														const CAStreamBasicDescription & inPrevFormat,
														const CAStreamBasicDescription & inNewFormat)
{
	ComponentResult result = AUBase::ChangeStreamFormat(inScope, inElement, inPrevFormat, inNewFormat);
	if (result == noErr)
	{
		if (inPrevFormat.mChannelsPerFrame != inNewFormat.mChannelsPerFrame)
			MaintainKernels();
	
		// for the moment this only dependency we know about
		// where a parameter's range may change is with the sample rate
		// and effects are only publishing parameters in the global scope!
		if (GetParamHasSampleRateDependency() && inPrevFormat.mSampleRate != inNewFormat.mSampleRate)
			PropertyChanged(kAudioUnitProperty_ParameterList, kAudioUnitScope_Global, 0);
	}
	
	return result;
}


// ____________________________________________________________________________
//
//	This method is called (potentially repeatedly) by ProcessForScheduledParams()
//	in order to perform the actual DSP required for this portion of the entire buffer
//	being processed.  The entire buffer can be divided up into smaller "slices"
//	according to the timestamps on the scheduled parameters...
//
ComponentResult	AUEffectBase::ProcessScheduledSlice(	void				*inUserData,
														UInt32				inStartFrameInBuffer,
														UInt32				inSliceFramesToProcess,
														UInt32				inTotalBufferFrames )
{
	ScheduledProcessParams	&sliceParams = *((ScheduledProcessParams*)inUserData);
	
	AudioUnitRenderActionFlags 	&actionFlags = *sliceParams.actionFlags;
	AudioBufferList 			&inputBufferList = *sliceParams.inputBufferList;
	AudioBufferList 			&outputBufferList = *sliceParams.outputBufferList;
	

	ComponentResult result = ProcessBufferLists(actionFlags, inputBufferList, outputBufferList, inSliceFramesToProcess );

	// we potentially just partially processed the buffers, so increment the data pointers to the next part of the
	// buffer to process
	for(unsigned int i = 0; i < inputBufferList.mNumberBuffers; i++ )
	{
		((Float32*)inputBufferList.mBuffers[i].mData) += inputBufferList.mBuffers[i].mNumberChannels * sizeof(Float32) * inSliceFramesToProcess;
		((Float32*)outputBufferList.mBuffers[i].mData) += outputBufferList.mBuffers[i].mNumberChannels * sizeof(Float32) * inSliceFramesToProcess;
	}
	
	return result;
}



// ____________________________________________________________________________
//

ComponentResult 	AUEffectBase::Render(	AudioUnitRenderActionFlags &ioActionFlags,
											const AudioTimeStamp &		inTimeStamp,
											UInt32						nFrames)
{
	if (!HasInput(0))
		return kAudioUnitErr_NoConnection;

	ComponentResult result = noErr;
	AUOutputElement *theOutput = GetOutput(0);	// throws if error

	result = PullInput(0 /* bus */, ioActionFlags, inTimeStamp, nFrames);
	
	if (result == noErr)
	{
		AUInputElement *input = GetInput(0);
		
		if (IsBypassEffect())
		{
			// leave silence bit alone
			input->CopyBufferContentsTo (theOutput->GetBufferList());
		}
		else
		{
			// this will read/write silence bit
			
			if(mParamList.size() == 0 )
			{
				result = ProcessBufferLists(ioActionFlags, input->GetBufferList(), theOutput->GetBufferList(), nFrames);
			}
			else
			{
				// deal with scheduled parameters...
				
				AudioBufferList &inputBufferList = input->GetBufferList();
				AudioBufferList &outputBufferList = theOutput->GetBufferList();
				
				ScheduledProcessParams processParams;
				processParams.actionFlags = &ioActionFlags;
				processParams.inputBufferList = &inputBufferList;
				processParams.outputBufferList = &outputBufferList;
	
				// divide up the buffer into slices according to scheduled params then
				// do the DSP for each slice (ProcessScheduledSlice() called for each slice)
				result = ProcessForScheduledParams(	mParamList,
													nFrames,
													&processParams );
	
				
				// fixup the buffer pointers to how they were before we started
				for(unsigned int i = 0; i < inputBufferList.mNumberBuffers; i++ )
				{
					((Float32*)inputBufferList.mBuffers[i].mData) -= inputBufferList.mBuffers[i].mNumberChannels * sizeof(Float32) * nFrames;
					((Float32*)outputBufferList.mBuffers[i].mData) -= outputBufferList.mBuffers[i].mNumberChannels * sizeof(Float32) * nFrames;
				}
			}
		}
	}
	
	return result;
}

OSStatus	AUEffectBase::ProcessBufferLists(
									AudioUnitRenderActionFlags &	ioActionFlags,
									const AudioBufferList &			inBuffer,
									AudioBufferList &				outBuffer,
									UInt32							inFramesToProcess)
{
	bool silentInput = (ioActionFlags & kAudioUnitRenderAction_OutputIsSilence) != 0;
	ioActionFlags |= kAudioUnitRenderAction_OutputIsSilence;
	bool ioSilence;
	UInt32 nBuffers, nChannels;
	if ((nBuffers=inBuffer.mNumberBuffers) != outBuffer.mNumberBuffers || nBuffers == 0)
		return kAudioUnitErr_FormatNotSupported;
	
	// call the kernels to handle either interleaved or deinterleaved
	if (nBuffers == 1) {
		// interleaved (or mono)
		int channel = 0;
		
		if ((nChannels=inBuffer.mBuffers[0].mNumberChannels) != outBuffer.mBuffers[0].mNumberChannels
			|| nChannels != mKernelList.size())
			return kAudioUnitErr_FormatNotSupported;
		
		for (KernelList::iterator it = mKernelList.begin(); it != mKernelList.end(); ++it, ++channel) {
			AUKernelBase *kernel = *it;
			
			if (kernel != NULL) {
				ioSilence = silentInput;
				
				// process each interleaved channel individually
				kernel->Process(
					(const Float32 *)inBuffer.mBuffers[0].mData + channel, 
					(Float32 *)outBuffer.mBuffers[0].mData + channel,
					inFramesToProcess,
					inBuffer.mBuffers[0].mNumberChannels,
					ioSilence);
					
				if (!ioSilence)
					ioActionFlags &= ~kAudioUnitRenderAction_OutputIsSilence;
			}
		}
		outBuffer.mBuffers[0].mDataByteSize = inFramesToProcess * channel * sizeof(Float32);
	} else {
		// deinterleaved
		if (nBuffers != mKernelList.size())
			return kAudioUnitErr_FormatNotSupported;
		
		const AudioBuffer *srcBuffer = inBuffer.mBuffers;
		AudioBuffer *destBuffer = outBuffer.mBuffers;
		
		for (KernelList::iterator it = mKernelList.begin(); it != mKernelList.end(); 
		++it, ++srcBuffer, ++destBuffer) {
			AUKernelBase *kernel = *it;
			
			if (kernel != NULL) {
				ioSilence = silentInput;
				
				kernel->Process(
					(const Float32 *)srcBuffer->mData, 
					(Float32 *)destBuffer->mData, 
					inFramesToProcess,
					1,
					ioSilence);
					
				if (!ioSilence)
					ioActionFlags &= ~kAudioUnitRenderAction_OutputIsSilence;
			}
			
			destBuffer->mDataByteSize = inFramesToProcess * sizeof(Float32);
		}
	}
	
	return noErr;
}

Float64		AUEffectBase::GetSampleRate()
{
	return GetOutput(0)->GetStreamFormat().mSampleRate;
}

UInt32		AUEffectBase::GetNumberOfChannels()
{
	return GetOutput(0)->GetStreamFormat().mChannelsPerFrame;
}

