/*
 *  TKnobControl.cpp
 *  NeuSynth
 *
 *  Created by Airy Andr on Sun Oct 20 2002.
 *  Copyright (c) 2002 Airy Andr. All rights reserved.
 *
 */

#include "TKnobControl.h"
#include "TImageCache.h"
#include "AUGUIUtilities.h"

//#define EXPERIMENT

using namespace AUGUI;

// -----------------------------------------------------------------------------
//	macros
// -----------------------------------------------------------------------------
//

// -----------------------------------------------------------------------------
//	TKnobControl constructor
// -----------------------------------------------------------------------------
//
TKnobControl::TKnobControl(
			   HIViewRef inControl )
:	TViewNoCompositingCompatible( inControl ), mImage(0), mBackImage(0),
mXOffset(0), mYOffset(0), mMinAngle(-135), mMaxAngle(135)
{
    //    ChangeAutoInvalidateFlags( kAutoInvalidateOnActivate | kAutoInvalidateOnEnable, 0 );
}

// -----------------------------------------------------------------------------
//	TKnobControl destructor
// -----------------------------------------------------------------------------
//	Clean up after yourself.
//
TKnobControl::~TKnobControl()
{
    if (mImage)
	CGImageRelease(mImage);
    if (mBackImage)
	CGImageRelease(mBackImage);
}

// -----------------------------------------------------------------------------
//	GetKind
// -----------------------------------------------------------------------------
//
ControlKind TKnobControl::GetKind()
{
    const ControlKind kMyKind = { 'airy', 'knob' };

    return kMyKind;
}


//-----------------------------------------------------------------------------------
//	Initialize
//-----------------------------------------------------------------------------------
//	The control is set up.  Do the last minute stuff that needs to be done like
//	installing EventLoopTimers.
//
OSStatus TKnobControl::Initialize(
				  TCarbonEvent&		inEvent )
{
#pragma unused( inEvent )
    TitleChanged();
    return noErr;
}
//-----------------------------------------------------------------------------------
//	ValueChanged
//-----------------------------------------------------------------------------------
//
void TKnobControl::ValueChanged()
{
    Invalidate();
}
//-----------------------------------------------------------------------------------
//	TitleChanged
//	The title format is
//	 	filename minAngle maxAngle xOffset xOffset
//	With :
//		<filename>.pgn and <filename>Back.png are the pict used by the control
//		minAngle and maxAngle the angle for min and max values (-135,135 default values)
//		xOffset and yOffset the offset of the rotating part from the back (0,0 default values)
//		Only the filename is mandatory
//-----------------------------------------------------------------------------------
void TKnobControl::TitleChanged()
{
    if (mImage)
	CGImageRelease(mImage);
    if (mBackImage)
	CGImageRelease(mBackImage);
    if ( mBundleRef )
    {
	CFStringRef fileName;
	CopyControlTitleAsCFString(GetViewRef(), &fileName);
	if ( fileName != NULL )
	{
	    char buffer[100];
	    char name[100];

	    char file[100];
	    float xOffset, yOffset, minAngle, maxAngle;
	    CFStringGetCString(fileName, buffer, 100, kCFStringEncodingASCII);
	    int nbVal = sscanf(buffer, "%s %f %f %f %f", file, &minAngle, &maxAngle, &xOffset, &yOffset);
	    if (nbVal < 2) minAngle = -135.;
	    if (nbVal < 3) maxAngle = 135.;
	    if (nbVal < 4) xOffset = 0.;
	    if (nbVal < 5) yOffset = 0.;
	    mMinAngle = minAngle;
	    mMaxAngle = maxAngle;
	    mXOffset = xOffset;
	    mYOffset = yOffset;

	    sprintf(name, "%s.png", file);
	    CFStringRef knob = CFStringCreateWithCStringNoCopy(0, name, kCFStringEncodingASCII, 0);
	    mImage = TImageCache::GetImage( mBundleRef, knob, NULL, NULL );

	    sprintf(name, "%sBack.png", file);
	    CFStringRef back = CFStringCreateWithCStringNoCopy(0, name, kCFStringEncodingASCII, 0);
		mBackImage = TImageCache::GetImage( mBundleRef, back, NULL, NULL );

	    CFRelease( fileName );
	}
    }
}

//-----------------------------------------------------------------------------------
//	DrawStaticBackground
//-----------------------------------------------------------------------------------
//!	Draw the static part (background) of the control
//
void TKnobControl::DrawStaticBackground(
										RgnHandle				inLimitRgn,
										CGContextRef			inContext,
										bool  inCompositing)
{
#pragma unused( inLimitRgn )
    TRect					bounds = Bounds();
    CGContextRef			context = inContext;

    if (mBackImage)
		HIViewDrawCGImage( context, &bounds, mBackImage);
}

//-----------------------------------------------------------------------------------
//	CompatibleDraw
//-----------------------------------------------------------------------------------
//!	The fun part of the control
//
void TKnobControl::CompatibleDraw(
				  RgnHandle				inLimitRgn,
				  CGContextRef			inContext,
				  bool  inCompositing)
{
#pragma unused( inLimitRgn )
    TRect					bounds = Bounds();
    CGContextRef			context = inContext;

    TRect knobBounds = bounds;

    if (mBackImage)
	HIViewDrawCGImage( context, &bounds, mBackImage);
    if (mImage) {
	float angle = (((float) (GetValue() - GetMinimum())/ (float) (GetMaximum()-GetMinimum())) * (mMaxAngle-mMinAngle) + mMinAngle);
	if (mBackImage)
	{
	    float scaleX = float(CGImageGetWidth(mImage)) / float(CGImageGetWidth(mBackImage));
	    float scaleY = float(CGImageGetHeight(mImage)) / float(CGImageGetHeight(mBackImage));
	    knobBounds.SetWidth(knobBounds.Width() * scaleX);
	    knobBounds.SetHeight(knobBounds.Height() * scaleY);
	}
	float knobWidth = knobBounds.Width();
	float knobHeight = knobBounds.Height();

	CGContextTranslateCTM(context, mXOffset + (bounds.MinX() + bounds.MaxX())/2., mYOffset + (bounds.MinY() + bounds.MaxY())/2.);
	CGContextRotateCTM( context, DegreesToRadians(angle) );
	knobBounds.MoveTo(-knobWidth/2, -knobHeight/2);
	HIViewDrawCGImage( context, &knobBounds, mImage);
    }
}

// -----------------------------------------------------------------------------
//	StillTracking
// -----------------------------------------------------------------------------
//!	Called during tracking.
//
OSStatus TKnobControl::StillTracking(TCarbonEvent&inEvent, HIPoint& from)
{
#ifdef EXPERIMENT
	const double kRatio = 250.;
#else
	const double kRatio = 50.;
#endif
    HIPoint mouse;
    float deltaX, deltaY;
    SInt32 mini, maxi;
    UInt32 modifiers;

    mini = GetMinimum();
    maxi = GetMaximum();

    inEvent.GetParameter<HIPoint>( kEventParamWindowMouseLocation, typeHIPoint, &mouse );
    inEvent.GetParameter( kEventParamKeyModifiers, &modifiers );
    ConvertToLocal(mouse);
    deltaX = mouse.x-from.x;
    deltaY = mouse.y-from.y;
    SInt32 curVal = GetValue();
    SInt32 val;
#ifdef EXPERIMENT
    TRect knobBounds = Bounds();
	knobBounds.MoveTo(0.,0.);
	if (from.x > knobBounds.Width()/2) {
		deltaY = -deltaY;
	}
	if (from.y > knobBounds.Height()/2) {
		deltaX = -deltaX;
	}
#endif

#if SUPPORT_CONTROL_MODIFIERS
    if (modifiers & HIGHEST_PRECISION_MODIFIER) {
		val = SInt32(rint(curVal+(deltaX-deltaY)*(maxi-mini)/(kRatio*400.)));
    } else if (modifiers & HIGH_PRECISION_MODIFIER) {
		val = SInt32(rint(curVal+(deltaX-deltaY)*(maxi-mini)/(kRatio*20.)));
    } else {
#endif
		val = SInt32(rint(curVal+(deltaX-deltaY)*(maxi-mini)/kRatio));
#if SUPPORT_CONTROL_MODIFIERS
    }
#endif
    if ( val > maxi) val = maxi;
    if ( val < mini) val = mini;
    if (val != curVal) {
		SetValue ( val );
		from = mouse;
    }
#ifdef EXPERIMENT
	from = mouse;
#endif
    return noErr;
}
