/*
 *  TViewNoCompositingCompatible.cpp
 *  NeuSynth
 *
 *  Created by Airy Andr on Wed Nov 06 2002.
 *  Copyright (c) 2002 Airy Andr. All rights reserved.
 *
 */

#include "TViewNoCompositingCompatible.h"
#include "TImageCache.h"
#include "TRect.h"
#include "AUGUIUtilities.h"

//-----------------------------------------------------------------------------------
//	constants (needed for the register/unregister stuff)
//-----------------------------------------------------------------------------------
//
static const EventTypeSpec kHIObjectEvents[] =
{
{ kEventClassHIObject, kEventHIObjectConstruct },
{ kEventClassHIObject, kEventHIObjectInitialize },
{ kEventClassHIObject, kEventHIObjectDestruct },
{ kEventClassHIObject, kEventHIObjectPrintDebugInfo }
};

CFMutableDictionaryRef TViewNoCompositingCompatible::mRegisteryCache = 0;
bool TViewNoCompositingCompatible::sWindowReady = true;
CFBundleRef TViewNoCompositingCompatible::sStaticBundle = 0;


TViewNoCompositingCompatible::TViewNoCompositingCompatible( HIViewRef inControl ) : TView( inControl ), mMouseEventHandler(0), mStaticBackground(0)
{
	mBundleRef = sStaticBundle;
	if (mBundleRef)
		CFRetain(mBundleRef);
};

//	Clean up after yourself.
TViewNoCompositingCompatible::~TViewNoCompositingCompatible()
{
    if (mBundleRef)
		CFRelease(mBundleRef);
	if (mStaticBackground)
		CFRelease(mStaticBackground);
}

void TViewNoCompositingCompatible::SetClassBundle(CFBundleRef sBundle)
{
    if (sStaticBundle)
		CFRelease(sBundle);

    if (sBundle)
		CFRetain(sBundle);
    sStaticBundle = sBundle;
}

//-----------------------------------------------------------------------------------
//	FlushStaticBackground
//-----------------------------------------------------------------------------------
//! Force the static background to be rebuild
//
void TViewNoCompositingCompatible::FlushStaticBackground()
{
	if (mStaticBackground) {
		CFRelease(mStaticBackground);
		mStaticBackground = 0;
		Invalidate();
		// We should also flush all of our subviews static backgrounds
	}
}

//-----------------------------------------------------------------------------------
//	mMouseEventHandler
//-----------------------------------------------------------------------------------
//!	Our static mouseDragged event handler proc.
//
OSStatus TViewNoCompositingCompatible::mouseDragged(TCarbonEvent&inEvent)
{
    return StillTracking(inEvent, mFromPoint);
}

//!	Our static mouseUp event handler proc.
OSStatus TViewNoCompositingCompatible::mouseUp(TCarbonEvent&inEvent)
{
    // Remove mouse handler...
    if (mMouseEventHandler) {
		RemoveEventHandler(mMouseEventHandler);
		mMouseEventHandler = 0;
    }

    OSStatus result = StopTracking(inEvent, mFromPoint);
	
	// create a control hit event to simulate a normal Track call
	if (result == noErr)
	{
		TCarbonEvent hitEvent(kEventClassControl, kEventControlHit);
		
		// get values from the in event to pass to the new event
		ControlPartCode hitPart = kControlNoPart;
		inEvent.GetParameter(kEventParamControlPart, &hitPart);
		UInt32 modifiers = 0;
		inEvent.GetParameter(kEventParamKeyModifiers, &modifiers);
		
		// set up parameters
		hitEvent.SetParameter<ControlRef>(kEventParamDirectObject, typeControlRef, GetViewRef());
		hitEvent.SetParameter(kEventParamControlPart, hitPart);
		hitEvent.SetParameter(kEventParamKeyModifiers, modifiers);
		
		// post to the current event queue
		hitEvent.PostToQueue(GetCurrentEventQueue());
	}
	return result;
}

static void releaseImageBuffer(void * info, const void* data, size_t dataSize)
{
	if (data)
		free(info); //const_cast<void*>(data));
}

//!	Draw background image into the context. Only the background part of the image hierarchy is taken in
//! account (we suppose no visible part of controls are overlapping)
void TViewNoCompositingCompatible::DrawBackgroundImage(CGContextRef context)
{
	
	HIViewRef superView = HIViewGetSuperview(GetViewRef());
	TViewNoCompositingCompatible *tview = 0;
	// Find our first superview which is a TViewNoCompositingCompatible
	while (superView && !tview) {
		OSErr err = GetControlData(superView, 0, theViewTag, sizeof(tview), &tview, 0);
		if (err != noErr) {
			superView = HIViewGetSuperview(superView);
		}
	}
	if (tview)
		tview->DrawBackgroundImage(context);
	DrawStaticBackground(NULL, context, IsCompositing());
}


//!	Create the offscreen CGImageRef containing the background of the control
//!	to speed up redraw. It does include the window background and the static
//!	part of the control, drawn by DrawStaticBackground method
void TViewNoCompositingCompatible::PrerenderStaticBackground()
{
	if (!sWindowReady) // Don't start compositing before the window is ready
		return;
	
	TRect bounds = Bounds();

	// get image info
	size_t width = (size_t)bounds.Width();
	size_t height = (size_t)bounds.Height();
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	size_t bitsPerComponent = 8;
	size_t bitsPerPixel = 32;
	size_t bytesPerRow = ((width * bitsPerPixel) + 7) / 8;
	CGImageAlphaInfo alphaInfo = kCGImageAlphaNoneSkipFirst;

	// create buffer and context
	size_t dataSize = bytesPerRow * height;
	void *buffer;
	void *data = AUGUI::malloc_aligned(dataSize, &buffer);
	if (!data)
		return;

	CGContextRef context = CGBitmapContextCreate(data, width, height, bitsPerComponent, bytesPerRow, colorSpace, alphaInfo);
	if (!context)
		return;
	
	// draw image into context
	CGContextTranslateCTM(context, 0, bounds.size.height);
	CGContextScaleCTM(context, 1.f, -1.f);
	CGContextTranslateCTM(context, -bounds.MinX(), -bounds.MinY());
	// We first fill the background with white color. This is for Controls not embedded in a pict control.
	// We suppose the host window is plain white.
	CGContextSetRGBFillColor(context, 0xFF, 0xFF, 0xFF, 0xFF);
	CGContextFillRect(context, bounds);
	DrawBackgroundImage(context);
	CGContextRelease(context);


	// create data provider for this image buffer
	CGDataProviderRef provider = CGDataProviderCreateWithData(buffer, data, dataSize, releaseImageBuffer);
	mStaticBackground = CGImageCreate(width, height, bitsPerComponent, bitsPerPixel, bytesPerRow, colorSpace, alphaInfo, provider, NULL, true,
								   kCGRenderingIntentDefault);
	CGDataProviderRelease(provider);

	CFRelease(colorSpace);
}


//! This is the event handler that is registered for the MouseUp and
//! MouseDragged events by the \c SetUpTracking member function.
pascal OSStatus TViewNoCompositingCompatible::MouseEventHandler(
																EventHandlerCallRef	inCallRef,
																EventRef			inEvent,
																void*				inUserData )
{
    OSStatus			result = eventNotHandledErr;
    TViewNoCompositingCompatible*				view = (TViewNoCompositingCompatible*) inUserData;
    TCarbonEvent		event( inEvent );

    switch ( event.GetClass() )
    {
		case kEventClassMouse:
			switch ( event.GetKind() )
			{
				case kEventMouseUp:
					result = view->mouseUp(event);
					break;

				case kEventMouseDragged:
					result = view->mouseDragged(event);
					break;
			}
			break;
    }
    return result;
}
//-----------------------------------------------------------------------------------
//	RegisterSubclass
//-----------------------------------------------------------------------------------
//!	This routine should be called by subclasses so they can be created as HIObjects.
//
OSStatus TViewNoCompositingCompatible::RegisterSubclass(
                                                        CFStringRef			inID,
                                                        ConstructProc		inProc )
{
    HIObjectClassRef classRef;

    OSStatus res = HIObjectRegisterSubclass( inID, kHIViewClassID, 0, ObjectEventHandler,
                                             GetEventTypeCount( kHIObjectEvents ), kHIObjectEvents, (void*) inProc, &classRef );
    if (res == noErr) {
        // Add (inId,classRef) in a dictionnary
        if (mRegisteryCache == 0) {
            mRegisteryCache = CFDictionaryCreateMutable(NULL, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
        }
        CFNumberRef number = CFNumberCreate(NULL, kCFNumberLongLongType, &classRef);
        if (CFDictionaryContainsKey(mRegisteryCache, inID))
            CFDictionaryReplaceValue(mRegisteryCache, inID, number);
        else
            CFDictionaryAddValue(mRegisteryCache, inID, number);
        CFRelease(number);
    }
    return res;
}
//-----------------------------------------------------------------------------------
//	UnRegisterSubclass
//-----------------------------------------------------------------------------------
//!	This routine should be called by subclasses to unregister themself.
//
OSStatus TViewNoCompositingCompatible::UnRegisterSubclass(CFStringRef			inID)
{
    OSStatus err = noErr;
    // Remove (inId,classRef) form the dictionnary and unregister classRef
    HIObjectClassRef classRef = 0;
    CFNumberRef number;
    if (CFDictionaryGetValueIfPresent(mRegisteryCache, inID, (const void **)&number)) {
		void* val;
		CFNumberGetValue(number, kCFNumberLongLongType, &val);
		classRef = HIObjectClassRef(val);
		if (classRef) err =  HIObjectUnregisterClass( classRef );
		CFDictionaryRemoveValue(mRegisteryCache, inID);
    }
    return err;
}

//-----------------------------------------------------------------------------------
//	SetUpTracking
//-----------------------------------------------------------------------------------
//!	Install handler for MouseUp and MouseDragged (for mouse tracking)
//
OSStatus TViewNoCompositingCompatible::SetUpTracking(TCarbonEvent& inEvent)
{
    static const EventTypeSpec kHIMouseEvents[] = {
	{ kEventClassMouse, kEventMouseUp },
	{ kEventClassMouse, kEventMouseDragged }
    };

    InstallEventHandler(
						GetWindowEventTarget(GetOwner()),
						MouseEventHandler,
						GetEventTypeCount(kHIMouseEvents),
						kHIMouseEvents,
						this,
						&mMouseEventHandler);
    inEvent.GetParameter<HIPoint>( kEventParamMouseLocation, typeHIPoint, &mFromPoint );
	if (!IsCompositing()) { // When called in a not compositing window : translate the coord to HIView local coords.
		TRect frame = Frame();
		mFromPoint.x -= frame.MinX();
		mFromPoint.y -= frame.MinY();
    }
    StartTracking(inEvent, mFromPoint);

    return noErr;
}

//-----------------------------------------------------------------------------------
//	Track
//-----------------------------------------------------------------------------------
//
OSStatus TViewNoCompositingCompatible::Track(TCarbonEvent& inEvent, ControlPartCode* outPart)
{
	*outPart = kControlNoPart;

	if (UseNonblockingTracking())
		return SetUpTracking(inEvent);
	else
		return eventNotHandledErr;
}


// To be overriden..
//-----------------------------------------------------------------------------------
//	StartTracking
//-----------------------------------------------------------------------------------
//!	Called at the beginning of tracking
//
OSStatus TViewNoCompositingCompatible::StartTracking(TCarbonEvent&inEvent, HIPoint& from)
{
    return noErr;
}

//-----------------------------------------------------------------------------------
//	StillTracking
//-----------------------------------------------------------------------------------
//!	Called whenever the mouse is moving during tracking
//
OSStatus TViewNoCompositingCompatible::StillTracking(TCarbonEvent&inEvent, HIPoint& from)
{
    return noErr;
}


//-----------------------------------------------------------------------------------
//	StopTracking
//-----------------------------------------------------------------------------------
//!	Called at the end of tracking. Our default implementation calls the HitTest()
//! method to determine which part of the control the mouse was over when the
//! button was released. This part code is set in \a inEvent.
//
OSStatus TViewNoCompositingCompatible::StopTracking(TCarbonEvent&inEvent, HIPoint& from)
{
	HIPoint mouse;
	inEvent.GetParameter<HIPoint>(kEventParamWindowMouseLocation, typeHIPoint, &mouse);
	ControlPartCode wherePart = HitTest(mouse);
	inEvent.SetParameter<ControlPartCode>(kEventParamControlPart, typeControlPartCode, wherePart);
    return noErr;
}

//-----------------------------------------------------------------------------------
//	Draw
//-----------------------------------------------------------------------------------
//!	Build a context for non-compositing window, draw the background and call the CompatibleDraw method
//
void TViewNoCompositingCompatible::Draw(
										RgnHandle				inLimitRgn,
										CGContextRef			inContext )
{
    CGContextRef			context = inContext;
    GrafPtr graphPtr = 0;
    bool compositing = (inContext != 0);

	TRect bounds = Bounds();

	if (!compositing) {
		SetPortWindowPort(GetOwner());
		Rect r = bounds;
		EraseRect(&r);
		if (mStaticBackground == 0) {
			PrerenderStaticBackground();
		}
	}
    if (!compositing) { // When called in a not compositing window
        graphPtr=GetWindowPort(GetOwner());
        
        Rect rect;
        GetPortBounds(graphPtr, &rect);
        
        QDBeginCGContext(graphPtr, &context);
        CGContextTranslateCTM(context, 0.f, float(rect.bottom-rect.top));
        CGContextScaleCTM(context, 1.f, -1.f);
        if (mStaticBackground)
            HIViewDrawCGImage(context, &bounds, mStaticBackground);
        else
            DrawStaticBackground(inLimitRgn, context, compositing);
            CGContextClipToRect(context, bounds);
    } else {
        DrawStaticBackground(inLimitRgn, context, compositing);
    }
        // Call the real draw routine
    CompatibleDraw(inLimitRgn, context, compositing);

    if (!compositing) {
		CGContextSynchronize ( context );
		QDEndCGContext (graphPtr, &context );
    }
}

//-----------------------------------------------------------------------------------
//	UnregisterAllClasses
//-----------------------------------------------------------------------------------
//!	Unregister all classes registred before
//
static void unregister(CFStringRef key, CFNumberRef value, void *context)
{
    void *val;
    CFNumberGetValue(value, kCFNumberLongLongType, &val);
    HIObjectClassRef classRef = HIObjectClassRef(val);
    if (classRef) {
		HIObjectUnregisterClass( classRef );
    }
}

void TViewNoCompositingCompatible::UnRegisterAllClasses()
{
    CFDictionaryApplyFunction(mRegisteryCache, (CFDictionaryApplierFunction)unregister, 0);
    CFDictionaryRemoveAllValues(mRegisteryCache);
}

//-----------------------------------------------------------------------------------
//	IsCompositing
//-----------------------------------------------------------------------------------
//!	Return true if the view is in a compositing window.
//
bool TViewNoCompositingCompatible::IsCompositing()
{
	WindowAttributes attributes;
    GetWindowAttributes(GetOwner(), &attributes);
    return attributes & kWindowCompositingAttribute;
}

//-----------------------------------------------------------------------------------
//	ConvertToLocal
//-----------------------------------------------------------------------------------
//!	Convert the point to view local (x,y).
//
void TViewNoCompositingCompatible::ConvertToLocal(HIPoint& thePoint)
{
	if (!IsCompositing()) { // When called in a not compositing window : convert to HIView local coords
		TRect frame = Frame();
		thePoint.x -= frame.MinX();
		thePoint.y -= frame.MinY();
    }
    HIViewConvertPoint(&thePoint, NULL, GetViewRef());
}

//-----------------------------------------------------------------------------------
//	ConvertToGlobal
//-----------------------------------------------------------------------------------
//!	Convert the local point to window (x,y).
//
void TViewNoCompositingCompatible::ConvertToGlobal(HIPoint& thePoint)
{
    HIViewConvertPoint(&thePoint, GetViewRef(), NULL);
}

//-----------------------------------------------------------------------------------
//	GetData
//-----------------------------------------------------------------------------------
//!	Get some data about the view
//! Only 'TVew' is supported. It is returning the C++ object.
//
OSStatus TViewNoCompositingCompatible::GetData(OSType inTag, ControlPartCode inPart, Size inSize, Size* outSize, void* inPtr)
{
	OSStatus err = noErr;

	switch (inTag)
	{
		case theViewTag:
			if (inSize == sizeof(TViewNoCompositingCompatible *))
			{
				if (inPtr)
					*(TViewNoCompositingCompatible **)inPtr = this;
				if (outSize)
					*outSize = sizeof(TViewNoCompositingCompatible *);
			}
			else
				err = errDataSizeMismatch;
			break;
		default:
			err = TView::GetData(inTag, inPart, inSize, outSize, inPtr);
	}
	return err;
}

//-----------------------------------------------------------------------------------
//	ControlHit
//-----------------------------------------------------------------------------------
//	The was hit.  Subclasses can overide to care about what part was hit.
//
OSStatus TViewNoCompositingCompatible::ControlHit(
						   ControlPartCode		inPart,
						   UInt32				inModifiers )
{
	OSStatus err = noErr;
	HICommandExtended theCommand;

	if (IsControlEnabled(GetViewRef())) {
		err = GetControlCommandID( GetViewRef(), &theCommand.commandID );
		require_noerr( err, CantGetCommandID );
		if ( theCommand.commandID != '\0\0\0\0') {
			EventRef commandEvent;
			err = CreateEvent( NULL, kEventClassCommand, kEventProcessCommand,
					  GetCurrentEventTime(), kEventAttributeNone, &commandEvent );
			require_noerr( err, CantCreateEvent );

			theCommand.attributes =  kHICommandFromControl;
			theCommand.source.control =  GetViewRef();
			err = SetEventParameter( commandEvent, kEventParamDirectObject,
							typeHICommand, sizeof( HICommand ), &theCommand );
			require_noerr( err, CantSetEventParameter );

			err = SendEventToEventTargetWithOptions( commandEvent, GetControlEventTarget(GetViewRef()) , kEventTargetSendToAllHandlers);
			return noErr;
		}
	}
CantSetEventParameter:
CantGetCommandID:
CantCreateEvent:
		return eventNotHandledErr;
}

// Asks your view to return what part of itself (if any) is hit by the point
// given to it. The point is in VIEW coordinates, so you should get the view
// srect to do bounds checking.
ControlPartCode TViewNoCompositingCompatible::HitTest(const HIPoint& inWhere)
{
	ControlPartCode part;
    // is the mouse on the button?
	if (CGRectContainsPoint(Bounds(), inWhere))
		part = kControlButtonPart;
	else
		part = kControlNoPart;

    return part;
}

//-----------------------------------------------------------------------------------
//	GetRegion
//-----------------------------------------------------------------------------------
//
OSStatus TViewNoCompositingCompatible::GetRegion(
							   ControlPartCode		inPart,
							   RgnHandle			outRgn )
{
    // For optimising redraw in compositing windows, subclass should return their real region (non transparent part)
    return TView::GetRegion(inPart, outRgn);
}




